#include <dlfcn.h>
#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>
#import <QuartzCore/QuartzCore.h>
#import <objc/Object.h>

void PrintStackTrace();
BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods);

static BOOL gSymbolicationLoaded = NO;
static BOOL loadSymbolication()
{
  if (gSymbolicationLoaded) {
    return YES;
  }
  // Needed for printing stack traces.  Available on OS X 10.5 and up.
  if (dlopen("/System/Library/PrivateFrameworks/Symbolication.framework/Symbolication", RTLD_LAZY)) {
    gSymbolicationLoaded = YES;
    Class VMUDebugMapExtractorClass = ::NSClassFromString(@"VMUDebugMapExtractor");
    SwizzleMethods(VMUDebugMapExtractorClass, @selector(debugMapExtractorWithMachOHeader:),
                   @selector(VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:), YES);
    SwizzleMethods([NSWindow class], @selector(toggleFullScreen:),
                   @selector(NSWindow_toggleFullScreen:), NO);
    SwizzleMethods([NSWindow class], @selector(setDelegate:),
                   @selector(NSWindow_setDelegate:), NO);
  }
  return gSymbolicationLoaded;
}

extern "C" void *NSPushAutoreleasePool();

static void *Hooked_NSPushAutoreleasePool()
{
  void *retval = NSPushAutoreleasePool();
  loadSymbolication();
  return retval;
}

static void alreadySwizzled(id self, SEL sel)
{
  NSLog(@"alreadySwizzled(): self %@\n", self);
}

@interface NSWindow (MethodSwizzling)
- (void)NSWindow_toggleFullScreen:(id)sender;
- (void)NSWindow_setDelegate:(id)delegate;
@end

@implementation NSWindow (MethodSwizzling)

- (void)NSWindow_toggleFullScreen:(id)sender
{
  NSLog(@"-[NSWindow toggleFullScreen:] start, self %@, sender %@", self, sender);
  [self NSWindow_toggleFullScreen:sender];
  NSLog(@"-[NSWindow toggleFullScreen:] finish, self %@, sender %@", self, sender);
}

- (void)NSWindow_setDelegate:(id)delegate
{
  //NSLog(@"-[NSWindow setDelegate:], self %@, delegate %@", self, delegate);
  if (delegate && ![delegate respondsToSelector:@selector(alreadySwizzled)]) {
    BOOL addMethodRetval = class_addMethod([delegate class], @selector(alreadySwizzled),
                                           (IMP) alreadySwizzled, "v@:");
    //NSLog(@"-[NSWindow setDelegate:], addMethodRetval %i", addMethodRetval);
    SwizzleMethods([delegate class], @selector(windowWillEnterFullScreen:),
                   @selector(NSWindowDelegate_windowWillEnterFullScreen:), NO);
    SwizzleMethods([delegate class], @selector(windowDidEnterFullScreen:),
                   @selector(NSWindowDelegate_windowDidEnterFullScreen:), NO);
    SwizzleMethods([delegate class], @selector(windowWillExitFullScreen:),
                   @selector(NSWindowDelegate_windowWillExitFullScreen:), NO);
    SwizzleMethods([delegate class], @selector(windowDidExitFullScreen:),
                   @selector(NSWindowDelegate_windowDidExitFullScreen:), NO);
    SwizzleMethods([delegate class], @selector(windowDidFailToEnterFullScreen:),
                   @selector(NSWindowDelegate_windowDidFailToEnterFullScreen:), NO);
    SwizzleMethods([delegate class], @selector(windowDidFailToExitFullScreen:),
                   @selector(NSWindowDelegate_windowDidFailToExitFullScreen:), NO);
    SwizzleMethods([delegate class], @selector(windowWillUseStandardFrame:defaultFrame:),
                   @selector(NSWindowDelegate_windowWillUseStandardFrame:defaultFrame:), NO);
  }
  [self NSWindow_setDelegate:delegate];
}

@end

@interface NSObject (NSWindowDelegateMethodSwizzling)
- (void)NSWindowDelegate_windowWillEnterFullScreen:(NSNotification*)notification;
- (void)NSWindowDelegate_windowDidEnterFullScreen:(NSNotification*)notification;
- (void)NSWindowDelegate_windowWillExitFullScreen:(NSNotification*)notification;
- (void)NSWindowDelegate_windowDidExitFullScreen:(NSNotification*)notification;
- (void)NSWindowDelegate_windowDidFailToEnterFullScreen:(NSWindow*)window;
- (void)NSWindowDelegate_windowDidFailToExitFullScreen:(NSWindow*)window;
- (NSRect)NSWindowDelegate_windowWillUseStandardFrame:(NSWindow*)window
                                         defaultFrame:(NSRect)newFrame;
//- (NSSize)NSWindowDelegate_window:(NSWindow*)window
//     willUseFullScreenContentSize:(NSSize)proposedSize;
//- (NSUInteger)NSWindowDelegate_window:(NSWindow*)window
//  willUseFullScreenPresentationOptions:(NSUInteger)proposedOptions;
@end

@implementation NSObject (NSWindowDelegateMethodSwizzling)

- (void)NSWindowDelegate_windowWillEnterFullScreen:(NSNotification*)notification
{
  NSLog(@"-[NSWindowDelegate windowWillEnterFullScreen:], self %@, notification %@",
        self, notification);
  [self NSWindowDelegate_windowWillEnterFullScreen:notification];
}

- (void)NSWindowDelegate_windowDidEnterFullScreen:(NSNotification*)notification
{
  NSLog(@"-[NSWindowDelegate windowDidEnterFullScreen:], self %@, notification %@",
        self, notification);
  [self NSWindowDelegate_windowDidEnterFullScreen:notification];
}

- (void)NSWindowDelegate_windowWillExitFullScreen:(NSNotification*)notification
{
  NSLog(@"-[NSWindowDelegate windowWillExitFullScreen:], self %@, notification %@",
        self, notification);
  [self NSWindowDelegate_windowWillExitFullScreen:notification];
}

- (void)NSWindowDelegate_windowDidExitFullScreen:(NSNotification*)notification
{
  NSLog(@"-[NSWindowDelegate windowDidExitFullScreen:], self %@, notification %@",
        self, notification);
  [self NSWindowDelegate_windowDidExitFullScreen:notification];
}

- (void)NSWindowDelegate_windowDidFailToEnterFullScreen:(NSWindow*)window
{
  NSLog(@"-[NSWindowDelegate windowDidFailToEnterFullScreen:], self %@, window %@",
        self, window);
  [self NSWindowDelegate_windowDidFailToEnterFullScreen:window];
}

- (void)NSWindowDelegate_windowDidFailToExitFullScreen:(NSWindow*)window
{
  NSLog(@"-[NSWindowDelegate windowDidFailToExitFullScreen:], self %@, window %@",
        self, window);
  [self NSWindowDelegate_windowDidFailToExitFullScreen:window];
}

- (NSRect)NSWindowDelegate_windowWillUseStandardFrame:(NSWindow*)window
                                         defaultFrame:(NSRect)newFrame
{
  NSRect retval = [self NSWindowDelegate_windowWillUseStandardFrame:window
                                                       defaultFrame:newFrame];
  NSLog(@"-[NSWindowDelegate windowWillUseStandardFrame:defaultFrame:], self %@, window %@, newFrame x %f y %f width %f height %f, returning x %f y %f width %f height %f",
        self, window, newFrame.origin.x, newFrame.origin.y, newFrame.size.width, newFrame.size.height,
        retval.origin.x, retval.origin.y, retval.size.width, retval.size.height);
  return retval;
}

@end

#pragma mark -

struct interpose_substitution {
  const void* replacement;
  const void* original;
};

#define INTERPOSE_FUNCTION(function) \
    { reinterpret_cast<const void*>(Hooked_##function), \
      reinterpret_cast<const void*>(function) }

__attribute__((used)) static const interpose_substitution substitutions[]
    __attribute__((section("__DATA, __interpose"))) = {
  INTERPOSE_FUNCTION(NSPushAutoreleasePool),
};

// What follows are definitions of the Symbolication interfaces we use to get
// stack traces.  The Symbolication framework is an undocumented, private
// framework available on OS X 10.5 and up.  It's used by Apple utilities
// such as dtrace.  http://www.cocoadev.com/index.pl?StackTraces has basic
// information on how to use the Symbolication framework.

@interface VMUAddressRange : NSObject <NSCoding>
{
}

@end

@interface VMUSymbol : VMUAddressRange <NSCopying>
{
}

- (NSString *)name;
- (NSString *)description;
- (id)text;

@end

@interface VMUSymbolOwner : NSObject <NSCopying>
{
}

- (NSString *)name;

@end

// We must override the debugMapExtractorWithMachOHeader method:  Otherwise
// the Symbolication framework can take a minute or more to initialize!!
// We don't need the information in the "debug maps" (Dwarf or dsym debugging
// information) to get simple stack traces.
@interface NSObject (VMUDebugMapExtractorMethodSwizzling)
+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader;
@end

@implementation NSObject (VMUDebugMapExtractorMethodSwizzling)

+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader
{
  return nil;
}

@end

@interface VMUSymbolicator : NSObject
{
}

+ (VMUSymbolicator *)symbolicatorForPid:(int)pid;
- (VMUSymbol *)symbolForAddress:(unsigned long long)address;
- (VMUSymbolOwner *)symbolOwnerForAddress:(unsigned long long)address;

@end

@interface VMUSymbolicator (PerformanceOptimization)
- (id)symbolOwnerNameForAddress:(unsigned long long)fp8;
@end

void PrintStackTrace()
{
  // +[NSThread callStackReturnAddress] only present on 10.5 and up.
  if (![NSThread respondsToSelector:@selector(callStackReturnAddresses)])
    return;
  if (!loadSymbolication())
    return;

  // Test for presence of Symbolication framework.
  Class symbolicatorClass = ::NSClassFromString(@"VMUSymbolicator");
  if (!symbolicatorClass)
    return;

  VMUSymbolicator *symbolicator = [symbolicatorClass symbolicatorForPid:getpid()];
  NSArray *addresses = [NSThread callStackReturnAddresses];
  NSUInteger count = [addresses count];
  for (NSUInteger i = 0; i < count; ++i) {
    NSNumber *item = (NSNumber *) [addresses objectAtIndex:i];
    VMUSymbol *symbol =
      [symbolicator symbolForAddress:[item unsignedLongLongValue]];
    NSString *ownerName = nil;
    if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
      ownerName =
        [symbolicator symbolOwnerNameForAddress:[item unsignedLongLongValue]];
    } else {
      VMUSymbolOwner *owner =
        [symbolicator symbolOwnerForAddress:[item unsignedLongLongValue]];
      ownerName = [owner name];
    }
    NSLog(@"    (%@) %@, %llu", ownerName, [symbol name], [item unsignedLongLongValue]);
  }
}

BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods)
{
  Method original = nil;
  Method posed = nil;

  if (classMethods) {
    original = class_getClassMethod(aClass, orgMethod);
    posed = class_getClassMethod(aClass, posedMethod);
  } else {
    original = class_getInstanceMethod(aClass, orgMethod);
    posed = class_getInstanceMethod(aClass, posedMethod);
  }

  if (!original || !posed)
    return NO;

  method_exchangeImplementations(original, posed);

  return YES;
}
